from unittest import mock
from .base import VaultHelperTest
import vaulthelpers


class VaultAuthenticationTest(VaultHelperTest):

    @mock.patch('os.getpid')
    @mock.patch('hvac.Client')
    def test_process_forking(self, Client, getpid):
        # Mock the PID, so that we can test process forking
        getpid.return_value = 1

        # Construct a mock Vault client
        mockClient = mock.MagicMock()
        Client.return_value = mockClient

        # Client should not have been created yet
        self.assertEqual(Client.call_count, 0)

        # Construct a client
        vaulthelpers.common.get_vault_auth().authenticated_client()

        # Client should have been created called once
        self.assertEqual(Client.call_count, 1)

        # Construct a client again
        vaulthelpers.common.get_vault_auth().authenticated_client()

        # Existing client should have been recycled, rather than recreated
        self.assertEqual(Client.call_count, 1)

        # Change the PID, simulating the process getting forked after a client already exists
        getpid.return_value = 2

        # Construct a client again
        vaulthelpers.common.get_vault_auth().authenticated_client()

        # Since the PID changed, the existing client should have been discarded and a new one created.
        self.assertEqual(Client.call_count, 2)


    def test_token_recycling(self):
        # Construct a client and record the token it gets after logging in
        vcl = vaulthelpers.common.get_vault_auth().authenticated_client()
        token_A = vcl.token

        # Construct a client again and get it's token.
        vcl = vaulthelpers.common.get_vault_auth().authenticated_client()
        token_B = vcl.token

        # Token should be reused, not regenerated.
        self.assertEqual(token_A, token_B)

        # Force the client to be destroyed, then create it again.
        vaulthelpers.common.reset_vault()
        vcl = vaulthelpers.common.get_vault_auth().authenticated_client()
        token_C = vcl.token

        # Token should be reused again, this time due to file caching.
        self.assertEqual(token_A, token_C)

        # Purge the file cache, force the client to be destroyed, then create the client again.
        vaulthelpers.common.get_vault_auth().purge_token_cache()
        vaulthelpers.common.reset_vault()
        vcl = vaulthelpers.common.get_vault_auth().authenticated_client()
        token_D = vcl.token

        # Token should be different this time.
        self.assertNotEqual(token_A, token_D)

        # Once more, destroy the client and get a token
        vaulthelpers.common.reset_vault()
        vcl = vaulthelpers.common.get_vault_auth().authenticated_client()
        token_E = vcl.token

        # Cache should have recycled token_D again
        self.assertEqual(token_D, token_E)



from botocore.exceptions import CredentialRetrievalError
from datetime import datetime, timedelta
from dateutil.parser import parse
from django.core.serializers.json import DjangoJSONEncoder
from . import common, utils
import portalocker
import logging
import boto3
import botocore
import pytz
import json
import hashlib
import os
import stat

logger = logging.getLogger(__name__)


class VaultProvider(botocore.credentials.CredentialProvider):
    """
    CredentialProvider plug-in for botocore to fetch and refresh credentials from Vault

    Roughly based off of ``botocore.credentials.EnvProvider``.
    """
    METHOD = 'vault'
    REFRESH_MARGIN = (60 * 10)  # Refresh credentials 10 minutes before they expire


    def __init__(self, path):
        """
        Initialize the Vault credentials provider.

        Arguments:
            path {string} -- Vault secret path to use to fetch AWS credentials
        """
        self.path = path


    @property
    def cache_filename(self):
        base_path = os.path.abspath(os.path.expanduser(common.VAULT_AWS_CACHE))
        path_hash = hashlib.md5(self.path.encode()).hexdigest().lower()
        return '{}-{}'.format(base_path, path_hash)


    @property
    def lock_filename(self):
        return '{}.lock'.format(self.cache_filename)


    def load(self):
        """Load AWS credentials from Vault

        Returns:
            {botocore.credentials.RefreshableCredentials} -- AWS Credentials object with built-in refresh callback
        """
        fetcher = self._create_credentials_fetcher()
        credentials = fetcher()
        return botocore.credentials.RefreshableCredentials(
            access_key=credentials['access_key'],
            secret_key=credentials['secret_key'],
            token=credentials['token'],
            expiry_time=parse(credentials['expiry_time']),
            refresh_using=fetcher,
            method=self.METHOD)


    def _create_credentials_fetcher(self):
        """Create a return a function which can be called to fetch AWS credentials form Vault

        Returns:
            {dictionary} -- Credentials dictionary with keys ``access_key``, ``secret_key``, ``token``, and ``expiry_time``
        """
        def fetch_credentials():
            # Obtain a lock file so prevent races between multiple processes trying to obtain tokens at the same time
            with portalocker.Lock(self.lock_filename, timeout=10):
                # Try to use cached credentials if at all possible
                credentials = self._read_credential_cache()
                if credentials:
                    return credentials

                # No cache, so obtain new credentials
                try:
                    vcl = common.get_vault_auth().authenticated_client()
                    result = vcl.read(self.path)
                except Exception as e:
                    utils.log_exception('Failed to load configuration from Vault at path {}.'.format(self.path))
                    raise CredentialRetrievalError(provider=self.METHOD, error_msg=str(e))
                expiry_time = datetime.utcnow().replace(tzinfo=pytz.utc)
                expiry_time += timedelta(seconds=result['lease_duration'])
                expiry_time -= timedelta(seconds=self.REFRESH_MARGIN)
                credentials = {
                    'access_key': result['data']['access_key'],
                    'secret_key': result['data']['secret_key'],
                    'token': result['data']['security_token'],
                    'expiry_time': expiry_time.strftime('%Y-%m-%d %H:%M:%S%z'),
                }
                self._write_credential_cache(credentials)
            return credentials

        return fetch_credentials


    def _read_credential_cache(self):
        # Try to read the cached credentials from the file system
        try:
            with open(self.cache_filename, 'r') as cache_file:
                data = json.load(cache_file)
        except OSError:
            return None

        # Parse the credentials expiration time
        expiry_time = None
        try:
            expiry_time = parse(data.get('expiry_time'))
        except ValueError:
            return None

        # If no expiry time was found, something went wrong. Return None
        if not expiry_time:
            return None

        # Check if the credentials are expired. If they are, return None
        refresh_threshold = (expiry_time - timedelta(seconds=self.REFRESH_MARGIN))
        if datetime.now(tz=pytz.UTC) > refresh_threshold:
            return None

        # Finally, return the cached data
        return data


    def _write_credential_cache(self, credentials):
        with open(self.cache_filename, 'w') as cache_file:
            json.dump(credentials, cache_file, cls=DjangoJSONEncoder)
        # Make the file only readable to the owner
        os.chmod(self.cache_filename, stat.S_IRUSR | stat.S_IWUSR)


    def purge_credential_cache(self):
        with portalocker.Lock(self.lock_filename, timeout=10):
            try:
                os.unlink(self.cache_filename)
            except FileNotFoundError:
                pass



class VaultSession(botocore.session.Session):
    def __init__(self, vault_path, *args, **kwargs):
        """Initialize a ``botocore`` session using credentials from Vault.

        Exactly the same as ``botocore.session.Session``, but uses AWS credentials from Vault.

        Arguments:
            vault_path {string} -- Vault secret path to use to fetch AWS credentials
        """
        self._vault_path = vault_path
        super().__init__(*args, **kwargs)


    def _register_credential_provider(self):
        """Registers a credential resolve instance with VaultProvider inserted as the most preferred credential provider."""
        def get_cred_resolver():
            credential_resolver = botocore.credentials.create_credential_resolver(self)
            vault_provider = VaultProvider(self._vault_path)
            credential_resolver.insert_before(botocore.credentials.EnvProvider.METHOD, vault_provider)
            return credential_resolver
        self._components.lazy_register_component('credential_provider', get_cred_resolver)



def init_boto3_credentials():
    """
    Configures boto3 to use AWS credentials from Vault
    """
    if not common.VAULT_AWS_PATH:
        logger.warning('Failed to load AWS credentials from Vault: missing AWS secret path.')
        return

    botocore_session = VaultSession(vault_path=common.VAULT_AWS_PATH)
    boto3.setup_default_session(botocore_session=botocore_session)



import sys
import logging

try:
    import sentry_sdk
except ImportError:
    sentry_sdk = None

logger = logging.getLogger(__name__)


def log_exception(msg):
    if sentry_sdk is not None:
        sentry_sdk.capture_exception()
        logger.error('{} Exception: {}'.format(msg, str(sys.exc_info()[1])))
    else:
        logger.debug('Could not report exception to Sentry because raven is not installed.')
        logger.exception(msg)



from storages.backends.s3boto3 import S3Boto3Storage as BaseS3Boto3Storage
import boto3


class S3Boto3Storage(BaseS3Boto3Storage):
    @property
    def connection(self):
        connection = getattr(self._connections, 'connection', None)
        if connection is None:
            self._connections.connection = boto3.DEFAULT_SESSION.resource('s3',
                region_name=self.region_name,
                use_ssl=self.use_ssl,
                endpoint_url=self.endpoint_url,
                config=self.config)
        return self._connections.connection



#!/usr/bin/env python
from setuptools import setup, find_packages, Distribution
import codecs
import os.path

# Make sure versiontag exists before going any further
Distribution().fetch_build_eggs('versiontag>=1.2.0')

from versiontag import get_version, cache_git_tag  # NOQA


packages = find_packages('src')

install_requires = [
    'Django>=1.11',
    'hvac>=0.3.0',
    'portalocker>=1.1.0',
    'pytz>=2017.2',
    'python-dateutil>=2.6.1'
]

extras_require = {
    'aws': [
        'boto3>=1.4.7',
        'botocore>=1.7.22',
    ],
    'database': [
        'dj-database-url>=0.4.2',
    ],
    'sentry': [
        'sentry-sdk>=0.5.5',
    ],
    'development': [
        'flake8>=3.3.0',
        'freezegun>=0.3.9',
        'psycopg2-binary>=2.7.1',
        'requests-mock>=1.4.0',
        'tox>=2.7.0',
        'versiontag>=1.2.0',
    ],
}



def fpath(name):
    return os.path.join(os.path.dirname(__file__), name)


def read(fname):
    return codecs.open(fpath(fname), encoding='utf-8').read()


cache_git_tag()

setup(
    name='django-vault-helpers',
    description="Helper functionality for obtaining secrets and credentials from Hashicorp Vault in a Django project",
    version=get_version(pypi=True),
    long_description=open('README.rst').read(),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: ISC License (ISCL)',
        'Operating System :: Unix',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    author='Craig Weber',
    author_email='crgwbr@gmail.com',
    url='https://gitlab.com/thelabnyc/django-vault-helpers',
    license='ISC',
    package_dir={'': 'src'},
    packages=packages,
    include_package_data=True,
    install_requires=install_requires,
    extras_require=extras_require,
)



from datetime import datetime, timedelta
from freezegun import freeze_time
from .base import VaultHelperTest
import vaulthelpers
import boto3
import pytz
import os
import requests_mock


class AWSCredentialsTest(VaultHelperTest):

    def setUp(self):
        super().setUp()
        boto3.DEFAULT_SESSION = None
        try:
            os.unlink('/tmp/.vault-aws-34ac139d5002477fce8d63447f0f2569')
            os.unlink('/tmp/.vault-aws-34ac139d5002477fce8d63447f0f2569.lock')
        except FileNotFoundError:
            pass
        vaulthelpers.aws.init_boto3_credentials()


    def mock_sts_creds(self, rmock, suffix):
        rmock.get('http://vault:8200/v1/aws/sts/vaulthelpers-sandbox', json={
            'lease_duration': 3600,
            'data': {
                'access_key': "access-key-{}".format(suffix),
                'secret_key': "secret-key-{}".format(suffix),
                'security_token': "security-token-{}".format(suffix),
            },
        })


    @requests_mock.mock(real_http=True)
    def test_obtain_sts_creds(self, rmock):
        self.mock_sts_creds(rmock, 'A')

        # Should fetch credentials
        creds = boto3._get_default_session().get_credentials()
        self.assertEqual(creds.access_key, "access-key-A")
        self.assertEqual(creds.secret_key, "secret-key-A")
        self.assertEqual(creds.token, "security-token-A")

        self.mock_sts_creds(rmock, 'B')

        # Should used cached credentials
        creds = boto3._get_default_session().get_credentials()
        self.assertEqual(creds.access_key, "access-key-A")
        self.assertEqual(creds.secret_key, "secret-key-A")
        self.assertEqual(creds.token, "security-token-A")


    @requests_mock.mock(real_http=True)
    def test_renew_sts_creds(self, rmock):
        now = datetime.now(tz=pytz.UTC)
        self.mock_sts_creds(rmock, 'A')

        # Should fetch credentials
        with freeze_time(now):
            creds = boto3._get_default_session().get_credentials()
            self.assertEqual(creds.access_key, "access-key-A")
            self.assertEqual(creds.secret_key, "secret-key-A")
            self.assertEqual(creds.token, "security-token-A")
            self.assertFalse(creds.refresh_needed())

        self.mock_sts_creds(rmock, 'B')

        # Should fetch credentials again, since the time changed
        with freeze_time(now + timedelta(hours=2)):
            creds = boto3._get_default_session().get_credentials()
            self.assertTrue(creds.refresh_needed())
            self.assertEqual(creds.access_key, "access-key-B")
            self.assertEqual(creds.secret_key, "secret-key-B")
            self.assertEqual(creds.token, "security-token-B")


    @requests_mock.mock(real_http=True)
    def test_sts_creds_auto_refresh(self, rmock):
        now = datetime.now(tz=pytz.UTC)
        self.mock_sts_creds(rmock, 'A')

        # Should fetch credentials
        with freeze_time(now):
            creds = boto3._get_default_session().get_credentials()
            self.assertEqual(creds.access_key, "access-key-A")
            self.assertEqual(creds.secret_key, "secret-key-A")
            self.assertEqual(creds.token, "security-token-A")
            self.assertFalse(creds.refresh_needed())

        self.mock_sts_creds(rmock, 'B')

        # Should fetch credentials again, since the time changed
        with freeze_time(now + timedelta(hours=2)):
            self.assertEqual(creds.access_key, "access-key-B")
            self.assertEqual(creds.secret_key, "secret-key-B")
            self.assertEqual(creds.token, "security-token-B")



from django.apps.config import AppConfig
from django.core.exceptions import ImproperlyConfigured
from django.db.backends.signals import connection_created
from . import aws  # NOQA
from . import common  # NOQA
from . import database  # NOQA
from . import utils
import os

default_app_config = 'vaulthelpers.VaultHelpersAppConfig'



def client():
    vault_auth = common.get_vault_auth()
    if not vault_auth:
        return
    vcl = vault_auth.authenticated_client()
    return vcl



class VaultHelpersAppConfig(AppConfig):
    name = 'vaulthelpers'

    def ready(self):
        # Register DB credential fetching code so that we can fetch credentials from Vault before
        # attempting to connect to the database.
        if common.VaultAuthenticator.has_envconfig():
            from django.conf import settings
            found = False
            for k, db in settings.DATABASES.items():
                if isinstance(db, database.DjangoAutoRefreshDBCredentialsDict):
                    found = True
            if found:
                database.monkeypatch_django()

        # Register SET_ROLE signal handler for the standard PostgreSQL database wrapper
        from django.db.backends.postgresql.base import DatabaseWrapper as PostgreSQLDatabaseWrapper
        connection_created.connect(database.set_role_connection, sender=PostgreSQLDatabaseWrapper)

        # Register SET_ROLE signal handler for the PostGIS database wrapper
        try:
            from django.contrib.gis.db.backends.postgis.base import DatabaseWrapper as PostGISDatabaseWrapper
            connection_created.connect(database.set_role_connection, sender=PostGISDatabaseWrapper)
        except (ImportError, ImproperlyConfigured):  # This exception will get thrown if the GDAL C libraries aren't installed.
            pass



class EnvironmentConfig(object):
    def __init__(self, path, kv_version=1):
        self.path = path
        self.config = {}
        try:
            vcl = client()
            if vcl:
                self.config = vcl.read(self.path).get('data', {})
        except Exception:
            utils.log_exception('Failed to load configuration from Vault at path {}.'.format(path))
        # Adjust for K/V API version
        if kv_version == 2:
            self.config = self.config.get('data', {})


    def get(self, name, default=None):
        value = self.config.get(name)
        if value:
            return value
        return os.environ.get(name, default)


    def __getitem__(self, name):
        if name in self.config:
            return self.config[name]
        if name in os.environ:
            return os.environ[name]
        raise KeyError(name)



from datetime import datetime, timedelta
from dateutil.parser import parse
from django.db.backends.base import base as django_db_base
from django.db import utils as django_db_utils
from django.core.serializers.json import DjangoJSONEncoder
from requests.exceptions import RequestException
from hvac.exceptions import InvalidRequest
from .exceptions import VaultCredentialProviderError
from . import common, utils
import logging
import dj_database_url
import pytz
import warnings
import portalocker
import hashlib
import os
import json
import stat

set_role_warning_given = False

_operror_types = ()
_operror_types += (django_db_utils.OperationalError,)
try:
    import psycopg2
except ImportError:
    pass
else:
    _operror_types += (psycopg2.OperationalError,)

try:
    import sqlite3
except ImportError:
    pass
else:
    _operror_types += (sqlite3.OperationalError,)

try:
    import MySQLdb
except ImportError:
    pass
else:
    _operror_types += (MySQLdb.OperationalError,)

logger = logging.getLogger(__name__)


DEFAULT_GRACE_SECONDS = (60 * 10)



class DatabaseCredentialProvider(object):
    def __init__(self, secret_path):
        self.secret_path = secret_path
        self._creds = None
        self._lease_id = None
        self._lease_expires = None


    @property
    def username(self):
        if self._creds is None:
            self.refresh_creds()
        return self._creds["username"]


    @property
    def password(self):
        if self._creds is None:
            self.refresh_creds()
        return self._creds["password"]


    @property
    def cache_filename(self):
        base_path = os.path.abspath(os.path.expanduser(common.VAULT_DB_CACHE))
        path_hash = hashlib.md5(self.secret_path.encode()).hexdigest().lower()
        return '{}-{}'.format(base_path, path_hash)


    @property
    def lock_filename(self):
        return '{}.lock'.format(self.cache_filename)


    def reset_creds(self):
        self._creds = None
        self._lease_id = None
        self._lease_expires = None


    def refresh_creds(self, lease_grace_period=DEFAULT_GRACE_SECONDS):
        logger.info('Beginning database credential refresh. Obtaining lock file.')
        # Obtain a lock file so prevent races between multiple processes trying to obtain credentials at the same time
        with portalocker.Lock(self.lock_filename, timeout=10):
            logger.info('Obtained lock file.')
            # Try to use cached credentials if at all possible
            data = self._read_credential_cache(lease_grace_period)
            if data:
                self._creds = data['creds']
                self._lease_id = data['lease_id']
                self._lease_expires = data['lease_expiration']
                logger.info("Loaded cached Vault DB credentials from filesystem %s: lease_id=[%s], expires=[%s], username=[%s]",
                    self.secret_path,
                    self._lease_id,
                    self._lease_expires.isoformat(),
                    self._creds['username'])
                return

            # No cache, so obtain new credentials
            logger.info('Failed to load credentials from cache. Getting authenticated Vault client.')
            vcl = common.get_vault_auth().authenticated_client()
            try:
                result = vcl.read(self.secret_path)
            except RequestException as e:
                raise VaultCredentialProviderError(
                    "Unable to read credentials from path '{}' with request error: {}".format(self.secret_path, str(e)))

            if "data" not in result or "username" not in result["data"] or "password" not in result["data"]:
                raise VaultCredentialProviderError(
                    "Read dict from Vault path {} did not match expected structure (data->{username, password}): %s".format(self.secret_path, str(result)))

            self._creds = result['data']
            self._lease_id = result['lease_id']
            self._lease_expires = datetime.now(tz=pytz.UTC) + timedelta(seconds=result['lease_duration'])
            self._write_credential_cache(self._creds, self._lease_id, self._lease_expires)
        logger.info("Loaded new Vault DB credentials from %s: lease_id=[%s], expires=[%s], username=[%s]",
            self.secret_path,
            self._lease_id,
            self._lease_expires.isoformat(),
            self._creds['username'])


    def refresh_creds_if_needed(self, lease_grace_period=DEFAULT_GRACE_SECONDS):
        logger.info('Determining is database credential refresh is needed. grace_period=[%s]', lease_grace_period)

        refresh = False
        # If we have no credentials at all, refresh the credentials.
        if self._creds is None:
            logger.info('Database credential refresh is needed because self._creds is None')
            refresh = True

        # If theres less than {lease_grace_period} seconds left in the lease, refresh the credentials.
        now = datetime.now(tz=pytz.UTC)
        graceful_expires = None
        if self._lease_expires is not None:
            graceful_expires = (self._lease_expires - timedelta(seconds=lease_grace_period))
        if graceful_expires is not None and now >= graceful_expires:
            logger.info('Database credential refresh is needed because self._lease_expires is within grace period. now=[%s], expires=[%s]',
                now,
                self._lease_expires)
            refresh = True

        # If lease got revoked, refresh the credentials.
        lease_ttl = self.fetch_lease_ttl()
        if lease_ttl <= lease_grace_period:
            logger.info('Database credential refresh is needed because lease_ttl is within grace period. lease_ttl=[%s], grace_period=[%s]',
                lease_ttl,
                lease_grace_period)
            refresh = True

        # If needed, refresh.
        if refresh:
            logger.info('Database credential refresh is needed. lease_ttl=[%s], grace_period=[%s]', lease_ttl, lease_grace_period)
            self.refresh_creds(lease_grace_period)
        else:
            logger.info('Database credential refresh is not needed. lease_ttl=[%s], grace_period=[%s]', lease_ttl, lease_grace_period)
        return


    def fetch_lease_ttl(self):
        client = common.get_vault_auth().authenticated_client()
        params = { "lease_id": self._lease_id }
        try:
            resp = client.adapter.put('/v1/sys/leases/lookup', json=params).json()
        except InvalidRequest as e:
            logger.info('Failed to fetch lease TTL from Vault. Assuming lease is expire. lease_id=[%s], error=[%s]', self._lease_id, e)
            return 0
        ttl = resp.get('data', {}).get('ttl', 0)
        logger.info('Fetched lease ID from Vault. lease_id=[%s], ttl=[%s]', self._lease_id, ttl)
        return ttl


    def _read_credential_cache(self, lease_grace_period):
        # Try to read the cached credentials from the file system
        try:
            with open(self.cache_filename, 'r') as cache_file:
                data = json.load(cache_file)
        except OSError as e:
            logger.info('Failed to read database credential cache from disk. error=[%s]', e)
            return None

        # Parse the credentials expiration time
        try:
            data['lease_expiration'] = parse(data.get('lease_expiration'))
        except ValueError as e:
            logger.info('Failed to read database credential cache because lease_expiration is invalid. error=[%s]', e)
            return None

        # If no expiry time was found, something went wrong. Return None
        if not data['lease_expiration']:
            logger.info('Failed to read database credential cache because lease_expiration is missing.')
            return None

        # Check if the credentials are expired. If they are, return None
        now = datetime.now(tz=pytz.UTC)
        refresh_threshold = (data['lease_expiration'] - timedelta(seconds=lease_grace_period))
        if now > refresh_threshold:
            logger.info('Failed to read database credential cache because cached credentials are expired. now=[%s], expired=[%s]', now, refresh_threshold)
            return None

        # Finally, return the cached data
        logger.info('Returning cached dataabase credentials. expires=[%s]', refresh_threshold)
        return data


    def _write_credential_cache(self, creds, lease_id, lease_expiration):
        data = {
            'creds': creds,
            'lease_id': lease_id,
            'lease_expiration': lease_expiration,
        }
        with open(self.cache_filename, 'w') as cache_file:
            json.dump(data, cache_file, cls=DjangoJSONEncoder)
        # Make the file only readable to the owner
        os.chmod(self.cache_filename, stat.S_IRUSR | stat.S_IWUSR)


    def purge_credential_cache(self):
        logger.info('Attempting to purge database credential cache. path=[%s]', self.cache_filename)
        with portalocker.Lock(self.lock_filename, timeout=10):
            try:
                os.unlink(self.cache_filename)
            except FileNotFoundError:
                logger.info('Failed to purge Database credential cache because cache file was not found. path=[%s]', self.cache_filename)
                pass



class DjangoAutoRefreshDBCredentialsDict(dict):
    def __init__(self, provider, *args, **kwargs):
        self._provider = provider
        super().__init__(*args, **kwargs)


    def refresh_credentials(self):
        lease_grace_period = self.get('OPTIONS', {}).get('vault_lease_grace_period', DEFAULT_GRACE_SECONDS)
        self._provider.refresh_creds_if_needed(lease_grace_period)
        self["USER"] = self._provider.username
        self["PASSWORD"] = self._provider.password


    def reset_credentials(self):
        self._provider.reset_creds()
        self["USER"] = None
        self["PASSWORD"] = None


    def purge_credential_cache(self):
        self._provider.purge_credential_cache()


    def __str__(self) -> str:
        return "DjangoAutoRefreshDBCredentialsDict(%s)" % super().__str__()


    def __repr__(self) -> str:
        return "DjangoAutoRefreshDBCredentialsDict(%s)" % super().__repr__()



def get_config(extra_config={}):
    """Load database configuration from Vault.

    Keyword Arguments:
        extra_config {dict} -- Extra keys for the returned configuration dictionary (default: {{}})

    Returns:
        {dictionary} -- Django database configuration
    """
    db_config = dj_database_url.config()
    db_config.update({
        'SET_ROLE': common.DATABASE_OWNERROLE,
    })
    db_config.update(extra_config)

    if not common.VAULT_DATABASE_PATH:
        logger.warning('Failed to load DB configuration from Vault: missing database secret path.')
        return db_config

    vault_creds = DatabaseCredentialProvider(common.VAULT_DATABASE_PATH)

    try:
        db_config.update({
            'USER': vault_creds.username,
            'PASSWORD': vault_creds.password,
        })
    except Exception:
        utils.log_exception('Failed to load configuration from Vault at path {}.'.format(common.VAULT_DATABASE_PATH))
        return db_config

    return DjangoAutoRefreshDBCredentialsDict(vault_creds, db_config)



def monkeypatch_django():
    def ensure_connection_with_retries(self):
        if self.connection is not None and self.connection.closed:
            logger.info("Failed database connection detected")
            self.connection = None

        if self.connection is None:
            with self.wrap_database_errors:
                try:
                    # Try to connect
                    self.connect()
                except Exception as e:
                    # See if this is a known error type or not
                    if not isinstance(e, _operror_types):
                        logger.warning("Database connection failed, but not due to a known error. errors=[%s]", e)
                        raise

                    # Get the max number of retry attempts
                    max_retries = self.settings_dict.get('OPTIONS', {}).get('vault_connection_retries', 3)

                    # If the max retry count has been exceeded, raise the error
                    if hasattr(self, "_vault_retries") and self._vault_retries >= max_retries:
                        logger.error("Retrying with new credentials from Vault didn't help. errors=[%s]", e)
                        raise

                    # Try to refresh Vault credentials and attempt another connection
                    logger.info("Database connection failed. Refreshing credentials from Vault.")
                    if not hasattr(self.settings_dict, 'refresh_credentials'):
                        logger.info("Installed vaulthelpers database connection settings_dict")
                        self.settings_dict = get_config(self.settings_dict)

                    # If we've already retried once, purge the cache and try again
                    if hasattr(self, "_vault_retries") and self._vault_retries >= 1:
                        logger.info("Purging credential cache before refreshing credentials from Vault.")
                        self.settings_dict.purge_credential_cache()

                    # Refresh the credentials from Vault and re-connect
                    self.settings_dict.refresh_credentials()
                    if not hasattr(self, "_vault_retries"):
                        self._vault_retries = 0
                    self._vault_retries += 1
                    self.ensure_connection()
                else:
                    # After a successful connection, reset the retry count back down to 0
                    self._vault_retries = 0

    logger.info("Installed vaulthelpers database connection helper into BaseDatabaseWrapper")
    django_db_base.BaseDatabaseWrapper.ensure_connection = ensure_connection_with_retries



def set_role_connection(sender, connection, **kwargs):
    global set_role_warning_given
    role = None
    if "set_role" in connection.settings_dict:
        role = connection.settings_dict["set_role"]
    elif "SET_ROLE" in connection.settings_dict:
        role = connection.settings_dict["SET_ROLE"]

    if role:
        connection.cursor().execute("SET ROLE %s", (role, ))
    else:
        if not set_role_warning_given:
            warnings.warn("Value for SET_ROLE is missing from settings.DATABASE")
            set_role_warning_given = True
